# frozen_string_literal: true

module ActsAsLiving::MethodsDefiner
  def self.call(klass)
    klass.class_eval do
      extend ClassMethods
      include InstanceMethods

      @stage_keys.each(&method(:define_stage_queries))
      @phases.each(&method(:define_phase_queries))
    end
  end

  module InstanceMethods
    def stage_after(stage)
      self.class.stage_after(stage)
    end

    def stages
      self.class.stages
    end

    def stage_before(stage)
      self.class.stage_before(stage)
    end

    def to_next_stage
      update(stage: next_stage)
    end

    def to_next_stage!
      update!(stage: next_stage)
    end

    def next_stage
      stage_after(stage) if stage
    end

    def to_previous_stage
      update(stage: previous_stage)
    end

    def to_previous_stage!
      update!(stage: previous_stage)
    end

    def previous_stage
      stage_before(stage)
    end

    def dying?
      self.class.dying?(stage)
    end

    def first_stage?
      self.class.first_stage?(stage)
    end

    def dead?
      self.class.death == stage
    end

    def dead_or_dying?
      dead? || dying?
    end

    def klass_phases_with_ranges
      self.class.phases_with_ranges
    end

    def klass_phases_for(stage)
      self.class.phases_for(stage)
    end

    def klass_stages
      self.class.stages
    end

    def locked?(&block)
      return unless block

      @locked_on.to_set.intersect? [stage, stage_was].to_set
    end

    def phases
      klass_phases_for(stage)
    end

    def phase_changed?
      klass_phases_for(stage) != klass_phases_for(stage_was)
    end

    def phases_started
      klass_phases_for(stage) - klass_phases_for(stage_was)
    end

    def phases_ended
      klass_phases_for(stage_was) - klass_phases_for(stage)
    end
  end

  module ClassMethods
    def alive_stages
      stages.except(@death).keys
    end

    def stage_keys
      @stage_keys
    end

    def stages_after(stage)
      return [] if dying?(stage)

      stages[stage_after(stage)..]
    end

    def stages_before(stage)
      return [] if first_stage?(stage)

      index = stage_keys.find_index(stage)
      stage_keys[0...index]
    end

    def stage_after(stage)
      return if dying?(stage)

      index = stage_keys.find_index(stage)
      stage_keys[index + 1]
    end

    def stage_before(stage)
      return if first_stage?(stage)

      index = stage_keys.find_index(stage)
      stage_keys[index - 1]
    end

    def last_stage
      (stage_keys - %i[cancelled]).last
    end

    def last_stage?(stage)
      last_stage.to_s == stage.to_s
    end

    def first_stage
      (stage_keys - %i[cancelled]).first
    end

    def first_stage?(stage)
      first_stage.to_s == stage.to_s
    end

    def death
      @death
    end

    def phases_with_ranges
      phases.map(&method(:to_phase_with_range)).to_h
    end

    def to_phase_with_range(phase, delimiter)
      [phase, (stages_enum[delimiter.first]..stages_enum[delimiter.last])]
    end

    def phases
      @phases
    end

    def phases_for(stage)
      phases_with_ranges.keys.select do |phase|
        phases_with_ranges[phase].include? stages[stage]
      end
    end

    def define_phase_queries(phase, delimiters)
      define_method("#{phase}?") do
        if delimiters.length == 1
          klass_stages[stage] == klass_stages[delimiters]
        else
          klass_stages[stage] >= klass_stages[delimiters.first] &&
            klass_stages[stage] <= klass_stages[delimiters.second]
        end
      end

      define_method("pre_#{phase}?") do
        klass_stages[stage] < klass_stages[delimiters.first] unless cancelled?
      end

      define_method("past_#{phase}?") do
        klass_stages[stage] > klass_stages[delimiters.last] || cancelled?
      end
    end

    def define_stage_queries(stage_key)
      define_method("stage_changed_to_#{stage_key}?") do
        stage_changed? && stage_was.to_s == stage_key.to_s
      end

      define_method("stage_saved_to_#{stage_key}?") do
        saved_change_to_stage? && stage.to_s == stage_key.to_s
      end

      define_method("pre_#{stage_key}?") do
        klass_stages[stage] < klass_stages[stage_key] unless cancelled?
      end

      define_method("past_#{stage_key}?") do
        klass_stages[stage] > klass_stages[stage_key] || cancelled?
      end

      define_method("cancelled_from_#{stage_key}?") do
        stage == 'cancelled' && stage_was.to_s == stage_key.to_s
      end
    end
  end
end
